"""Classification task."""

from .core import Classification, TaskFewshotExample, _TaskInferenceMode, _TaskPromptSignature, _TaskResult

__all__ = ["Classification", "TaskFewshotExample", "_TaskInferenceMode", "_TaskResult", "_TaskPromptSignature"]
